@@ -19,19 +19,21 @@ default_params = {
     'plugin_utility_path': PLUGIN_UTILITY_DIR,
     'plugin_utility_example_path': os.path.join(PLUGIN_UTILITY_DIR, 'Example'),
     'plugin_config_filename': 'plugin_config.json',
-    'plugin_search_dir': os.getenv('PLUGIN_DIR', os.path.join(PROJECT_DIR, 'plugins')),
     'media_path': os.path.join(PROJECT_DIR, 'background', 'media'),
     'document_path': os.path.join(PROJECT_DIR, 'background', 'media', 'documents'),
+    'plugin_search_dir': os.getenv('PLUGINS_DIR', os.path.join(PROJECT_DIR, 'plugins')),
+    'plugin_download_dir': os.getenv('PLUGIN_DOWNLOAD_DIR', os.path.join(PROJECT_DIR, 'downloads', 'plugins')),
 }
 
-from bootutils import us, ab, env_global
-from plugin_utility.iplugin import iPlugin
+from general_utility import utils as us
+from general_utility.appbase import ab
+
 from databar_utility.databar import Databar
 
-ENV = env_global(PROJECT_DIR)
+from plugin_utility.iplugin import iPlugin
 
-TRY_LOCAL_PLUGINS = ENV("TRY_LOCAL_PLUGINS", default=True)
-PIP_MIRRORS_URL = ENV("PIP_MIRRORS_URL", default=None)
+TRY_LOCAL_PLUGINS = os.getenv("TRY_LOCAL_PLUGINS", 'True')
+PIP_MIRRORS_URL = os.getenv("PIP_MIRRORS_URL", None)
 
 #######################################################################################
 
@@ -59,8 +61,8 @@ def get_timer_from_url(url):
     if (job_config is None):
         return None
 
-    timer_entry = us.check_attribute(json, 'values.entry.timers', 'timers')
-    values = us.check_attribute(json, ('values.%s' % timer_entry), 'now')
+    timer_entry = us.check_attribute(job_config, 'values.entry.timers', 'timers')
+    values = us.check_attribute(job_config, ('values.%s' % timer_entry), 'now')
     if (isinstance(values, str)):
         return values
 
@@ -92,7 +94,7 @@ def plugin_handle(url):
 
 class Plugin_Runner(object):
 
-    def __init__(self, plugin_url, local_params = None):
+    def __init__(self, config, local_params = None):
         self.__plugins = []
 
         if (local_params is not None):
@@ -102,16 +104,16 @@ class Plugin_Runner(object):
 
         self.params = {}
 
-        type = us.get_type(plugin_url)
+        type = us.get_type(config)
 
         if (type.lower() == 'dir'):
-            self.load_from_dirs(plugin_url)
+            self.load_from_dirs(config)
         elif (type.lower() == 'file'):
-            self.load_from_file(plugin_url)
+            self.load_from_file(config)
         elif (type.lower() == 'http') or (type.lower == 'https'):
-            self.load_from_url(plugin_url)
+            self.load_from_url(config)
         elif (type.lower() == 'dict') or (type.lower() == 'json'):
-            self.load_from_json(params)
+            self.load_from_json(config)
         pass
 
         #####################################################################
@@ -122,12 +124,10 @@ class Plugin_Runner(object):
         if (us.path_exists(config_file)):
             # 按照配置文件来加载
             import json
-            plugin_config_jaon = json.load(open(config_file, 'r', encoding='UTF-8'))
-            testdata = plugin_config_jaon['testdata']
-        else:
-            pass
+            plugin_config_json = json.load(open(config_file, 'r', encoding='UTF-8'))
+            testdata = plugin_config_json['testdata']
+            self.run(testdata)
 
-        self.run(testdata)
         pass
 
     def run(self, params=None): 
@@ -138,6 +138,22 @@ class Plugin_Runner(object):
 
         return self.project_infos_handle(params)
 
+    def on_scripts(self, scripts, moment='START'):
+        for script in scripts:
+            if (us.check_attribute(script, 'moment', 'START').upper() == moment.upper()):
+                value = us.check_attribute(script, 'script', None)
+
+                import tempfile
+                script_filename = tempfile.mktemp(suffix='.sh')
+                with open(script_filename, "wb") as f:
+                    f.write(value.replace("\r", "").encode())
+                f.close
+
+                ab.print_log('执行脚本.(%s : %s)' % (moment, script_filename))
+                os.system('chmod +x %s' % script_filename)
+                os.system(script_filename)
+            pass
+
     def project_infos_handle(self, params): # 项目数据流入口
         if (params is None):
             ab.print_log("project_infos_handle params is None, so exit!")
@@ -173,6 +189,10 @@ class Plugin_Runner(object):
                                 (params['last_result'], params['last_error']))
                 break
 
+        scripts = us.check_attribute(params, "scripts", None)
+        if (scripts is not None):
+            self.on_scripts(scripts, 'END')
+            
         return params
 
     #######################################################################################
@@ -185,18 +205,26 @@ class Plugin_Runner(object):
         self.__plugins.append(plugin)
 
     def addPlugins(self, plugins):
+        if (plugins is None):
+            return
         for plug in plugins:
             self.addPlugin(plug)
 
     def delPlugin(self, plugin):
+        if (plugins is None):
+            return
         if plug in self.__plugins:
             self.__plugins.remove(plugin)
 
     def delPlugins(self, plugins):
+        if (plugins is None):
+            return
         for plug in plugins:
             self.delPlugin(plug)
 
     def getPlugin(self, name = None):
+        if (plugins is None):
+            return
         ab.print_log("get plugin: %s" % name)
         for plugin in self.__plugins:
             if ((name is not None) and (plugin.name.upper() == name.upper())):
@@ -274,34 +302,45 @@ class Plugin_Runner(object):
     def load_from_json(self, json_obj):
 
         #####################################################################
-
-        id = json_obj.pop('id', -1)
-        meta = json_obj.pop('meta', None)
+        
+        id = us.check_attribute(json_obj, 'id', -1)
+        meta = us.check_attribute(json_obj, 'meta', None)
 
         #####################################################################
 
-        values = json_obj.pop('values', None)
-        self.params['protocol_version'] = values.pop('protocol_version', '2.0')
-        self.params['timers'] = values.pop('timers', None)
-        self.params['entry'] = values.pop('entry', None)
+        values = us.check_attribute(json_obj, 'values', None)
+        self.params['protocol_version'] = us.check_attribute(values, 'protocol_version', '2.0')
+        self.params['timers'] = us.check_attribute(values, 'timers', None)
+        self.params['entry'] = us.check_attribute(values, 'entry', None)
 
         #####################################################################
 
         key_name = us.check_attribute(self.params['entry'], 'base', 'base')
-        base = values.pop(key_name, {})
+        base = us.check_attribute(values, key_name, {})
         self.params[key_name] = base
         self.params[key_name] = {"id": id, "meta": meta, 'local_setting': default_params}
 
         #####################################################################
 
-        plugins = values.pop(us.check_attribute(self.params['entry'], 'engine', 'plugins'), None)
+        key_name = us.check_attribute(self.params['entry'], 'scripts', 'scripts')
+        scripts = us.check_attribute(values, key_name, None)
+        self.params[key_name] = scripts
+        self.params['next'] = key_name
+
+        # 脚本执行
+        if (scripts is not None):
+            self.on_scripts(scripts, 'START')
+
+        #####################################################################
+
+        plugins = us.check_attribute(values, us.check_attribute(self.params['entry'], 'engine', 'plugins'), None)
         plugins = plugins_loader(plugins, self.params['protocol_version'])
         self.addPlugins(plugins.get_values())
 
         #####################################################################
 
         key_name = us.check_attribute(self.params['entry'], 'settings', 'settings')
-        setting = values.pop(key_name, None)
+        setting = us.check_attribute(values, key_name, None)
         setting = setting_loader(setting, self.params['protocol_version'])
         self.params[key_name] = setting.get_values()
 
@@ -316,14 +355,14 @@ class Plugin_Runner(object):
         #####################################################################
 
         key_name = us.check_attribute(self.params['entry'], 'databars', 'databars')
-        databar = values.pop(key_name, None)
+        databar = us.check_attribute(values, key_name, None)
         self.params[key_name] = databar
         self.params['next'] = key_name
 
         #####################################################################
 
         key_name = us.check_attribute(self.params['entry'], 'renders', 'renders')
-        render = values.pop(key_name, None)
+        render = us.check_attribute(values, key_name, None)
         render = base_loader(render, self.params['protocol_version'])
         self.params[key_name] = render.get_values()
 
@@ -332,7 +371,7 @@ class Plugin_Runner(object):
         #####################################################################
 
         key_name = us.check_attribute(self.params['entry'], 'processors', 'processors')
-        processors = values.pop(key_name, [])
+        processors = us.check_attribute(values, key_name, [])
         self.params[key_name] = processors
 
         pass
@@ -346,7 +385,7 @@ class base_loader():
 
     def get_values(self):
         return self.__values
-
+    
     def load_2_1(self, values):
         if (values is None):
             return None
@@ -371,6 +410,12 @@ class base_loader():
         elif (isinstance(values, dict)):
             return values
 
+    def load_2_3_1(self, values):
+        return self.load_2_2(values)
+
+    def load_2_3_2(self, values):
+        return self.load_2_2(values)
+
     def load(self, values):
         fun_name = ('load_%s' % self.__protocol_version.replace('.', '_'))
         if us.obj_hasattr(self, fun_name):
@@ -411,14 +456,22 @@ class plugins_loader(base_loader):
     def load_2_2(self, values):
         return self.load_2_1(values)
 
-    def load_plugins(self, plugins, plugin_search_dir = default_params['plugin_search_dir']):
+    def load_2_3_1(self, values):
+        return self.load_2_1(values)
+
+    def load_2_3_2(self, values):
+        return self.load_2_1(values)
+
+    def load_plugins(self, plugins, 
+                            plugin_search_dir = default_params['plugin_search_dir'], 
+                            plugin_download_dir = default_params['plugin_download_dir']):
         if (plugins is None):
             return
 
         plugins_list = []
         ab.print_log("load plugins obj from plugins list %s." % plugins.items())
         for (name, obj) in plugins.items():
-            if (TRY_LOCAL_PLUGINS) and (us.path_exists(os.path.join(plugin_search_dir, name))):
+            if (TRY_LOCAL_PLUGINS.upper() == "TRUE") and (us.path_exists(os.path.join(plugin_search_dir, name))):
                 ab.print_log("plugin search dir: %s" % (plugin_search_dir))
                 url =  os.path.join(plugin_search_dir, name)
 
@@ -441,12 +494,10 @@ class plugins_loader(base_loader):
                         url = host + url
 
                     ab.print_log("plugins %s is not exists on local, and load from web addr : %s" % (name, url))
-                elif (url_type not in 'https'):
-                    url = os.path.join(plugin_dir, url)
                 else:
                     pass
                     
-            plugin = Plugin_Object(url)
+            plugin = Plugin_Object(url, plugin_download_dir)
             if (plugin != None):
                 plugins_list.append(plugin)
 
@@ -455,8 +506,9 @@ class plugins_loader(base_loader):
 #######################################################################################
 
 class Plugin_Object(object):
-    def __init__(self, plugin_url):
+    def __init__(self, plugin_url, plugin_download_dir = None):
         self.obj = None
+        self.plugin_dir = plugin_download_dir
         if (os.path.isdir(plugin_url)):
             self.obj = self.load_from_dir(plugin_url)
         elif (os.path.isfile(plugin_url)):
@@ -464,7 +516,7 @@ class Plugin_Object(object):
         else:
             scheme = us.get_scheme_from_url(plugin_url)
             if (scheme != None) and (scheme.lower() in 'https'):
-                self.obj = self.load_from_url(plugin_url)
+                self.obj = self.load_from_url(plugin_url, self.plugin_dir)
 
     def __iter__(self):
         return iter(self.obj)
@@ -484,27 +536,14 @@ class Plugin_Object(object):
             ab.print_log("plugin config file is not exists (%s)" % config_file, "error")
             return None
 
-        if (us.path_exists(os.path.join(plugin_url, 'init.sh')) == True):
-            initfile = os.path.join(plugin_url, 'init.sh')
-            os.system('chmod +x %s' % initfile)
-            os.system(initfile)
-
-        if (us.path_exists(os.path.join(plugin_url, 'requirements.txt')) == True):
-            filename = os.path.join(plugin_url, 'requirements.txt')
-            pip_install_cmd = (f'python -m pip install -r %s' % (filename))
-            if (PIP_MIRRORS_URL is not None):
-                pip_install_cmd = (f'%s -i %s' % (pip_install_cmd, PIP_MIRRORS_URL))
-
-            ab.print_log(pip_install_cmd)
-            os.system(pip_install_cmd)
-            # from pip._internal import main as pip_main # 插件依赖包安装
-            # pip_main(['install', '-i', mirrors_url, '-r', filename])
-
         ab.print_log('load plugin obj from %s' % config_file)
         import json
         self.config = json.load(open(config_file, 'r', encoding='UTF-8'))
         start_file = self.config['base']['start']
         start_name = start_file[:-3]
+        
+        us.initialize(plugin_url, '')
+
         try:
             acquire_lock()
             fh, filename, desc = find_module(start_name, [plugin_url])
@@ -569,11 +608,18 @@ class Plugin_Object(object):
 
         _, _, name, suffix = us.get_path_name_suffix_from_filepath(plugin_url)
         if (suffix != '') and (suffix.lower() == '.zip'): # 暂时只处理没有后缀或者后缀为zip的文件
-            import tempfile
-            plugin_dir = tempfile.mkdtemp()
+            if (self.plugin_dir is None):
+                import tempfile
+                plugin_dir = tempfile.mkdtemp()
+            else:
+                plugin_dir = self.plugin_dir
+            
             plugin_dir = os.path.join(plugin_dir, name)
-            filecount = us.extract_file(plugin_url, plugin_dir)
-            ab.print_log("extract file count: (%d)" % filecount)
+            if (not us.path_exists(plugin_dir)):
+                us.check_path_exists(plugin_dir)
+
+                filecount = us.extract_file(plugin_url, plugin_dir)
+                ab.print_log("extract file count: (%d)" % filecount)
             return self.load_from_dir(plugin_dir)
         elif (suffix != '') and (suffix.lower() == '.py'):
             return self.load_plugin(plugin_url)
@@ -584,9 +630,9 @@ class Plugin_Object(object):
         # 临时文件咋删除？
         pass
 
-    def load_from_url(self, plugin_url):
+    def load_from_url(self, plugin_url, download_dir):
         ab.print_log('load plugin from url: %s' % plugin_url)
-        pulgin_filepath = us.get_filepath_from_url(plugin_url)
+        pulgin_filepath = us.get_filepath_from_url(plugin_url, download_dir)
         if (pulgin_filepath  == None):
             return
 
